from aws_adfs import commands


class TestVersion:

    def test_returns_version(self):
        # given

        # when
        try:
            result = commands.cli(['--version'])
            assert False
        except SystemExit as e:
            # then
            assert e.code == 0


import click
import lxml.etree as ET

from fido2.client import ClientData, ClientError, U2fClient
from fido2.hid import CtapHidDevice
try:
    from fido2.pcsc import CtapPcscDevice
except ImportError:
    CtapPcscDevice = None

import logging
import json
import re

from threading import Event, Thread

try:
    # Python 3
    from urllib.parse import urlparse, parse_qs
    import queue
except ImportError:
    # Python 2
    from urlparse import urlparse, parse_qs
    import Queue as queue

from . import roles_assertion_extractor

_headers = {
    'Accept-Language': 'en',
    'User-Agent': 'Mozilla/5.0 (Windows NT 6.1; WOW64; Trident/7.0; rv:11.0) like Gecko',
    'Content-Type': 'application/x-www-form-urlencoded; charset=UTF-8',
    'Accept': 'text/plain, */*; q=0.01',
}


def extract(html_response, ssl_verification_enabled, u2f_trigger_default, session):
    """
    this strategy is based on description from: https://duo.com/docs/duoweb
    :param response: raw http response
    :param html_response: html result of parsing http response
    :return:
    """

    duo_host = _duo_host(html_response)
    duo_request_signature = _duo_request_signature(html_response)
    roles_page_url = _action_url_on_validation_success(html_response)

    click.echo("Sending request for authentication", err=True)
    (sid, preferred_factor, preferred_device, u2f_supported), initiated = _initiate_authentication(
        duo_host,
        duo_request_signature,
        roles_page_url,
        session,
        ssl_verification_enabled
    )
    if initiated:
        click.echo("Waiting for additional authentication", err=True)

        rq = queue.Queue()
        auth_count = 0
        if u2f_supported:
            # Trigger U2F authentication
            auth_count += 1
            t = Thread(
                target=_perform_authentication_transaction,
                args=(
                    duo_host,
                    sid,
                    preferred_factor,
                    preferred_device,
                    True,
                    session,
                    ssl_verification_enabled,
                    rq,
                )
            )
            t.daemon = True
            t.start()

        if u2f_trigger_default or not u2f_supported:
            # Trigger default authentication (call or push) concurrently to U2F
            auth_count += 1
            t = Thread(
                target=_perform_authentication_transaction,
                args=(
                    duo_host,
                    sid,
                    preferred_factor,
                    preferred_device,
                    False,
                    session,
                    ssl_verification_enabled,
                    rq,
                )
            )
            t.daemon = True
            t.start()

        while "Wait for responses":
            auth_signature = rq.get()
            auth_count -= 1
            if auth_signature != "cancelled":
                break
            if auth_count < 1:
                click.echo("All authentication methods cancelled, aborting.")
                exit(-2)

        click.echo('Going for aws roles', err=True)
        return _retrieve_roles_page(
            roles_page_url,
            _context(html_response),
            session,
            ssl_verification_enabled,
            '{}:{}'.format(auth_signature, _app(duo_request_signature)),
        )

    return None, None, None


def _perform_authentication_transaction(duo_host, sid, preferred_factor, preferred_device, use_u2f, session, ssl_verification_enabled, rq):
    if (preferred_factor is None or preferred_device is None) and not use_u2f:
        click.echo("No default authentication method configured.", err=True)
        rq.put("cancelled")
        return

    transaction_id = _begin_authentication_transaction(
        duo_host,
        sid,
        preferred_factor,
        preferred_device,
        use_u2f,
        session,
        ssl_verification_enabled
    )

    transaction_id =_verify_authentication_status(
        duo_host,
        sid,
        transaction_id,
        session,
        ssl_verification_enabled,
    )
    if transaction_id == "cancelled":
        rq.put("cancelled")
    else:
        rq.put(
            _authentication_result(
                duo_host, sid, transaction_id, session, ssl_verification_enabled
            )
        )


def _context(html_response):
    context_query = './/input[@id="context"]'
    element = html_response.find(context_query)
    return element.get('value')


def _retrieve_roles_page(roles_page_url, context, session, ssl_verification_enabled,
                         signed_response):
    logging.debug('context: {}'.format(context))
    logging.debug('sig_response: {}'.format(signed_response))

    response = session.post(
        roles_page_url,
        verify=ssl_verification_enabled,
        headers=_headers,
        allow_redirects=True,
        data={
            'AuthMethod': 'DuoAdfsAdapter',
            'Context': context,
            'sig_response': signed_response,
        }
    )
    logging.debug(u'''Request:
            * url: {}
            * headers: {}
        Response:
            * status: {}
            * headers: {}
            * body: {}
        '''.format(roles_page_url, response.request.headers, response.status_code, response.headers,
                   response.text))

    if response.status_code != 200:
        raise click.ClickException(
            u'Issues during redirection to aws roles page. The error response {}'.format(
                response
            )
        )

    # Save session cookies to avoid having to repeat MFA on each login
    session.cookies.save(ignore_discard=True)

    html_response = ET.fromstring(response.text, ET.HTMLParser())
    return roles_assertion_extractor.extract(html_response)


def _authentication_result(
        duo_host,
        sid,
        duo_transaction_id,
        session,
        ssl_verification_enabled
):
    status_for_url = "https://{}/frame/status".format(duo_host)
    response = session.post(
        status_for_url,
        verify=ssl_verification_enabled,
        headers=_headers,
        data={
            'sid': sid,
            'txid': duo_transaction_id
        }
    )
    logging.debug(u'''Request:
        * url: {}
        * headers: {}
    Response:
        * status: {}
        * headers: {}
        * body: {}
    '''.format(status_for_url, response.request.headers, response.status_code, response.headers,
               response.text))

    if response.status_code != 200:
        raise click.ClickException(
            u'Issues during retrieval of a code entered into '
            u'the device. The error response {}'.format(
                response
            )
        )

    json_response = response.json()
    if json_response['stat'] != 'OK':
        raise click.ClickException(
            u'There was an issue during retrieval of a code entered into the device.'
            u' The error response: {}'.format(
                response.text
            )
        )

    if json_response['response']['status_code'] != 'allow':
        raise click.ClickException(
            u'There was an issue during retrieval of a code entered into the device.'
            u' The error response: {}'.format(
                response.text
            )
        )
    result_url = response.json()['response']['result_url']
    duo_result_response = _load_duo_result_url(duo_host, result_url, sid, session, ssl_verification_enabled)
    auth_signature = duo_result_response.json()['response']['cookie']
    return auth_signature

def _load_duo_result_url(
        duo_host,
        result_url,
        sid,
        session,
        ssl_verification_enabled
):
    result_for_url = 'https://{}'.format(duo_host) + result_url
    response = session.post(
        result_for_url,
        verify=ssl_verification_enabled,
        headers=_headers,
        data={
            'sid': sid
        }
    )
    logging.debug(u'''Request:
        * url: {}
        * headers: {}
    Response:
        * status: {}
        * headers: {}
        * body: {}
    '''.format(result_for_url, response.request.headers, response.status_code, response.headers,
               response.text))

    if response.status_code != 200:
        raise click.ClickException(
            u'Issues when following the Duo result URL after '
            u'authentication. The error response {}'.format(
                response
            )
        )
    json_response = response.json()
    if json_response['stat'] != 'OK':
        raise click.ClickException(
            u'There was an issue when following the Duo result URL after authentication.'
            u' The error response: {}'.format(
                response.text
            )
        )
    return response

def _verify_authentication_status(duo_host, sid, duo_transaction_id, session,
                                  ssl_verification_enabled):
    responses = []
    while len(responses) < 10:
        status_for_url = "https://{}/frame/status".format(duo_host)
        response = session.post(
            status_for_url,
            verify=ssl_verification_enabled,
            headers=_headers,
            data={
                'sid': sid,
                'txid': duo_transaction_id
            }
        )
        logging.debug(u'''Request:
            * url: {}
            * headers: {}
        Response:
            * status: {}
            * headers: {}
            * body: {}
        '''.format(status_for_url, response.request.headers, response.status_code, response.headers,
                response.text))

        if response.status_code != 200:
            raise click.ClickException(
                u'Issues during second factor verification. The error response {}'.format(
                    response
                )
            )

        json_response = response.json()
        if json_response['stat'] != 'OK':
            raise click.ClickException(
                u'There was an issue during second factor verification. The error response: {}'.format(
                    response.text
                )
            )

        if json_response['response']['status_code'] not in ['answered', 'calling', 'pushed', 'u2f_sent']:
            raise click.ClickException(
                u'There was an issue during second factor verification. The error response: {}'.format(
                    response.text
                )
            )

        if json_response['response']['status_code'] in ['pushed', 'answered', 'allow']:
            return duo_transaction_id

        if json_response['response']['status_code'] == 'u2f_sent' and len(json_response['response']['u2f_sign_request']) > 0:
            u2f_sign_requests = json_response['response']['u2f_sign_request']

            # appId, challenge and session is the same for all requests, get them from the first
            u2f_app_id = u2f_sign_requests[0]['appId']
            u2f_challenge = u2f_sign_requests[0]['challenge']
            u2f_session_id = u2f_sign_requests[0]['sessionId']

            devices = list(CtapHidDevice.list_devices())
            if CtapPcscDevice:
                devices.extend(list(CtapPcscDevice.list_devices()))

            if not devices:
                click.echo("No FIDO U2F authenticator is eligible.")
                return "cancelled"

            threads = []
            u2f_response = {
                "sessionId": u2f_session_id
            }
            rq = queue.Queue()
            cancel = Event()
            for device in devices:
                t = Thread(
                    target=_u2f_sign,
                    args=(
                        device,
                        u2f_app_id,
                        u2f_challenge,
                        u2f_sign_requests,
                        duo_host,
                        sid,
                        u2f_response,
                        session,
                        ssl_verification_enabled,
                        cancel,
                        rq
                    )
                )
                t.daemon = True
                threads.append(t)
                t.start()

            # Wait for first answer
            return rq.get()

        responses.append(response.text)

    raise click.ClickException(
        u'There was an issue during second factor verification. The responses: {}'.format(
            responses
        )
    )


def _u2f_sign(device, u2f_app_id, u2f_challenge, u2f_sign_requests, duo_host, sid, u2f_response, session, ssl_verification_enabled, cancel, rq):
    click.echo("Activate your FIDO U2F authenticator now: '{}'".format(device), err=True)
    client = U2fClient(device, u2f_app_id)
    try:
        u2f_response.update(
                client.sign(
                u2f_app_id,
                u2f_challenge,
                u2f_sign_requests,
                event=cancel
            )
        )

        # Cancel the other U2F prompts
        cancel.set()
        
        click.echo("Got response from FIDO U2F authenticator: '{}'".format(device), err=True)
        rq.put(_submit_u2f_response(duo_host, sid, u2f_response, session, ssl_verification_enabled))
    except:
        pass
    finally:
        device.close()


_tx_pattern = re.compile("(TX\|[^:]+):APP.+")


def _tx(request_signature):
    m = _tx_pattern.search(request_signature)
    return m.group(1)


_app_pattern = re.compile(".*(APP\|[^:]+)")


def _app(request_signature):
    m = _app_pattern.search(request_signature)
    return m.group(1)


def _initiate_authentication(duo_host, duo_request_signature, roles_page_url, session,
                             ssl_verification_enabled):
    prompt_for_url = 'https://{}/frame/web/v1/auth'.format(duo_host)
    response = session.post(
        prompt_for_url,
        verify=ssl_verification_enabled,
        headers={
            'Host': duo_host,
            'User-Agent': "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/75.0.3770.100 Safari/537.36",
            'Accept': "text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8",
            'Accept-Language': "en-US,en;q=0.5",
            'Accept-Encoding': "gzip, deflate, br",
            'DNT': "1",
            'Upgrade-Insecure-Requests': "1",
            'Content-Type': "application/x-www-form-urlencoded",
        },
        allow_redirects=True,
        params={
            'tx': _tx(duo_request_signature),
            'parent': roles_page_url,
            'v': '2.3',
        },
        data={
            'parent': roles_page_url,
            'java_version': '',
            'flash_version': '',
            'screen_resolution_width': '1280',
            'screen_resolution_height': '800',
            'color_depth': '24',
        }
    )
    logging.debug(u'''Request:
        * url: {}
        * headers: {}
    Response:
        * status: {}
        * headers: {}
        * body: {}
    '''.format(prompt_for_url, response.request.headers, response.status_code, response.headers,
               response.text))

    if response.status_code != 200 or response.url is None:
        return (None, None, None), False

    o = urlparse(response.url)
    query = parse_qs(o.query)

    if 'sid' not in query:
        return (None, None, None), False

    sid = query['sid']
    html_response = ET.fromstring(response.text, ET.HTMLParser())
    preferred_factor = _preferred_factor(html_response)
    preferred_device = _preferred_device(html_response)
    u2f_supported = _u2f_supported(html_response)
    return (sid, preferred_factor, preferred_device, u2f_supported), True


def _preferred_factor(html_response):
    preferred_factor_query = './/input[@name="preferred_factor"]'
    element = html_response.find(preferred_factor_query)
    return element is not None and element.get('value') or None


def _preferred_device(html_response):
    preferred_device_query = './/input[@name="preferred_device"]'
    element = html_response.find(preferred_device_query)
    return element is not None and element.get('value') or None


def _u2f_supported(html_response):
    u2f_supported_query = './/input[@name="factor"][@value="U2F Token"]'
    elements = html_response.findall(u2f_supported_query)
    return len(elements) > 0


def _begin_authentication_transaction(duo_host, sid, preferred_factor, preferred_device, u2f_supported, session,
                                      ssl_verification_enabled):
    prompt_for_url = "https://{}/frame/prompt".format(duo_host)
    if u2f_supported:
        response = session.post(
            prompt_for_url,
            verify=ssl_verification_enabled,
            headers=_headers,
            data={
                'sid': sid,
                'factor': "U2F Token",
                'device': 'u2f_token',
                'post_auth_action': ''
            }
        )
    else:
        click.echo("Triggering default authentication method: '{}'".format(preferred_factor), err=True)
        response = session.post(
            prompt_for_url,
            verify=ssl_verification_enabled,
            headers=_headers,
            data={
                'sid': sid,
                'factor': preferred_factor,
                'device': preferred_device,
                'out_of_date': ''
            }
        )
    logging.debug(u'''Request:
        * url: {}
        * headers: {}
    Response:
        * status: {}
        * headers: {}
        * body: {}
    '''.format(prompt_for_url, response.request.headers, response.status_code, response.headers,
               response.text))

    if response.status_code != 200:
        raise click.ClickException(
            u'Issues during beginning of the authentication process. The error response {}'.format(
                response
            )
        )

    json_response = response.json()
    if json_response['stat'] != 'OK':
        raise click.ClickException(
            u'Cannot begin authentication process. The error response: {}'.format(response.text)
        )

    return json_response['response']['txid']


def _submit_u2f_response(duo_host, sid, u2f_response, session, ssl_verification_enabled):
    prompt_for_url = "https://{}/frame/prompt".format(duo_host)
    response = session.post(
        prompt_for_url,
        verify=ssl_verification_enabled,
        headers=_headers,
        data={
            'sid': sid,
            'device': 'u2f_token',
            'factor': "u2f_finish",
            'response_data': json.dumps(u2f_response),
        }
    )
    logging.debug(u'''Request:
        * url: {}
        * headers: {}
    Response:
        * status: {}
        * headers: {}
        * body: {}
    '''.format(prompt_for_url, response.request.headers, response.status_code, response.headers,
               response.text))

    if response.status_code != 200:
        raise click.ClickException(
            u'Issues during submitting U2F response for the authentication process. The error response {}'.format(
                response
            )
        )

    json_response = response.json()
    if json_response['stat'] != 'OK':
        raise click.ClickException(
            u'Cannot complete authentication process. The error response: {}'.format(response.text)
        )

    return json_response['response']['txid']


_duo_host_pattern = re.compile("'host': '([^']+)'")


def _duo_host(html_response):
    try:
        duo_host_query = './/form[@id="duo_form"]/following-sibling::script'
        element = html_response.xpath(duo_host_query)[0]
        m = _duo_host_pattern.search(element.text)
        return m.group(1)
    except:
        duo_host_query = './/input[@name="duo_host"]/@value'
        return html_response.xpath(duo_host_query)[0]

_duo_signature_pattern = re.compile("'sig_request': '([^']+)'")


def _duo_request_signature(html_response):
    try:
        duo_signature_query = './/form[@id="duo_form"]/following-sibling::script'
        element = html_response.xpath(duo_signature_query)[0]
        m = _duo_signature_pattern.search(element.text)
        return m.group(1)
    except:
        duo_host_query = './/input[@name="duo_sig_request"]/@value'
        return html_response.xpath(duo_host_query)[0]


def _action_url_on_validation_success(html_response):
    duo_auth_method = './/form[@id="options"]'
    element = html_response.find(duo_auth_method)
    return element.get('action')



import logging
import lxml.etree as ET

from . import account_aliases_fetcher
from . import _duo_authenticator as duo_auth
from . import _rsa_authenticator as rsa_auth
from . import _symantec_vip_access as symantec_vip_access
from . import _azure_mfa_authenticator as azure_mfa_auth
from . import _azure_cloud_mfa_authenticator as azure_cloud_mfa_auth
from . import html_roles_fetcher
from . import roles_assertion_extractor


def authenticate(config, username=None, password=None, assertfile=None):
    response, session = html_roles_fetcher.fetch_html_encoded_roles(
        adfs_host=config.adfs_host,
        adfs_cookie_location=config.adfs_cookie_location,
        ssl_verification_enabled=config.ssl_verification,
        adfs_ca_bundle=config.adfs_ca_bundle,
        provider_id=config.provider_id,
        username=username,
        password=password,
        sspi=config.sspi,
        u2f_trigger_default=config.u2f_trigger_default,
    )

    assertion = None
    aws_session_duration = None

    aggregated_principal_roles = None
    if response.status_code == 200:
        extract_strategy = _strategy(response, config, session, assertfile)

        principal_roles, assertion, aws_session_duration = extract_strategy()

        if assertion is None:
            logging.debug(u'''Cannot extract saml assertion from request's response. Re-authentication needed?:
                * url: {}
                * headers: {}
            Response:
                * status: {}
                * headers: {}
                * body: {}
            '''.format(
                response.url,
                response.request.headers,
                response.status_code,
                response.headers,
                response.text
            ))
            logging.error(u'Cannot extract saml assertion. Re-authentication needed?')
        else:
            aggregated_principal_roles = _aggregate_roles_by_account_alias(session,
                                                                           config,
                                                                           username,
                                                                           password,
                                                                           assertion,
                                                                           principal_roles)

    else:
        logging.debug(u'''Cannot extract roles from request's response:
                * url: {}
                * headers: {}
            Response:
                * status: {}
                * headers: {}
                * body: {}
            '''.format(
            response.url,
            response.request.headers,
            response.status_code,
            response.headers,
            response.text
        ))
        logging.error(u'Cannot extract roles from response')

    logging.debug(u'Roles along with principals found after authentication: {}'.format(aggregated_principal_roles))

    return aggregated_principal_roles, assertion, aws_session_duration


def _aggregate_roles_by_account_alias(session,
                                      config,
                                      username,
                                      password,
                                      assertion,
                                      principal_roles):
    account_aliases = account_aliases_fetcher.account_aliases(session, username, password, config.provider_id, assertion, config)
    aggregated_accounts = {}
    for (principal_arn, role_arn) in principal_roles:
        role_name = role_arn.split(':role/')[1]
        account_no = role_arn.split(':')[4]

        if account_no not in account_aliases:
            account_aliases[account_no] = account_no

        if account_aliases[account_no] not in aggregated_accounts:
            aggregated_accounts[account_aliases[account_no]] = {}
        aggregated_accounts[account_aliases[account_no]][role_arn] = {'name': role_name, 'principal_arn': principal_arn}
    return aggregated_accounts


def _strategy(response, config, session, assertfile=None):

    html_response = ET.fromstring(response.text, ET.HTMLParser())

    def _plain_extractor():
        def extract():
            return roles_assertion_extractor.extract(html_response)
        return extract

    def _duo_extractor():
        def extract():
            return duo_auth.extract(html_response, config.ssl_verification, config.u2f_trigger_default, session)
        return extract

    def _symantec_vip_extractor():
        def extract():
            return symantec_vip_access.extract(html_response, config.ssl_verification, session)
        return extract

    def _file_extractor():
        def extract():
            return roles_assertion_extractor.extract_file(assertfile)
        return extract

    def _rsa_auth_extractor():
        def extract():
            return rsa_auth.extract(html_response, config.ssl_verification, session)
        return extract

    def _azure_mfa_extractor():
        def extract():
            return azure_mfa_auth.extract(html_response, config.ssl_verification, session)
        return extract

    def _azure_cloud_mfa_extractor():
        def extract():
            return azure_cloud_mfa_auth.extract(html_response, config.ssl_verification, session)
        return extract

    if assertfile is None:
        chosen_strategy = _plain_extractor
    else:
        chosen_strategy = _file_extractor

    if _is_duo_authentication(html_response):
        chosen_strategy = _duo_extractor
    elif _is_symantec_vip_authentication(html_response):
        chosen_strategy = _symantec_vip_extractor
    elif _is_rsa_authentication(html_response):
        chosen_strategy = _rsa_auth_extractor
    elif _is_azure_mfa_authentication(html_response):
        chosen_strategy = _azure_mfa_extractor
    elif _is_azure_cloud_mfa_authentication(html_response):
        chosen_strategy = _azure_cloud_mfa_extractor

    return chosen_strategy()


def _is_duo_authentication(html_response):
    duo_auth_method = './/input[@id="authMethod"]'
    element = html_response.find(duo_auth_method)
    duo = element is not None
    duo = duo and element.get('value') == 'DuoAdfsAdapter'
    return duo

def _is_symantec_vip_authentication(html_response):
    auth_method = './/input[@id="authMethod"]'
    element = html_response.find(auth_method)
    if (
        element is not None
        and element.get('value') == 'SymantecVipAdapter'
    ) or (
        element is not None
        and element.get('value') == 'VIPAuthenticationProviderWindowsAccountName'
    ):
        return True

def _is_rsa_authentication(html_response):
    auth_method = './/input[@id="authMethod"]'
    element = html_response.find(auth_method)
    return (
        element is not None
        and element.get('value') == 'SecurIDAuthentication'
    )

def _is_azure_mfa_authentication(html_response):
    auth_method = './/input[@id="authMethod"]'
    element = html_response.find(auth_method)
    return (
        element is not None
        and element.get('value') == 'AzureMfaServerAuthentication'
    )

def _is_azure_cloud_mfa_authentication(html_response):
    auth_method = './/input[@id="authMethod"]'
    element = html_response.find(auth_method)
    return (
        element is not None
        and element.get('value') == 'AzureMfaAuthentication'
    )



import sys

from aws_adfs import commands


class TestClickParsesCommandLineArguments:

    def test_parses_help_argument(self):
        # when executes aws-adfs --help
        sys.argv = ['aws-adfs', '--help']
        try:
            commands.cli()
        except SystemExit as e:
            # displays help description
            assert e.code == 0



from aws_adfs import account_aliases_fetcher


def _aws_account(account_alias, account_no):
    return u'<div class="saml-account-name">Account: {} ({})</div>'.format(account_alias, account_no)


def _aws_account_without_alias(account_no):
    return u'<div class="saml-account-name">Account: {}</div>'.format(account_no)


def _account_page_response_text(response_text):
    response = type('', (), {})()
    response.request = type('', (), {})()
    response.request.headers = {}
    response.status_code = 'irrelevant'
    response.headers = {}
    response.text = response_text

    return response


def _account_page_response(accounts):
    return _account_page_response_text(u'''
    <html>
        <body>
            <div>
                <form>
                    <fieldset>
                        {}
                    </fieldset>
                </form>
            </div>
        </body>
    </html>
    '''.format('\n'.join([account for account in accounts]))
    )


def _failed_account_page_response():
    return _account_page_response_text(u'<html></html>')


class TestAccountAliasesFetcher:
    
    def test_returns_empty_account_dictionary_when_no_account_are_named(self):
        # given user with no aws accounts
        self.authenticated_session.post = lambda *args, **kwargs: _account_page_response([])

        # when gets account aliases via fetcher
        accounts = account_aliases_fetcher.account_aliases(self.authenticated_session,
                                                           self.irrelevant_username,
                                                           self.irrelevant_password,
                                                           self.irrelevant_auth_method,
                                                           self.authenticated_saml_response,
                                                           self.irrelevant_config)
        # then returns no accounts
        assert accounts == {}

    def test_returns_one_account_when_one_account_is_listed(self):
        # given user with no aws accounts
        account_no = '123'
        account_alias = 'single'
        self.authenticated_session.post = lambda *args, **kwargs: _account_page_response([_aws_account(account_alias, account_no)])

        # when gets account aliases via fetcher
        accounts = account_aliases_fetcher.account_aliases(self.authenticated_session,
                                                           self.irrelevant_username,
                                                           self.irrelevant_password,
                                                           self.irrelevant_auth_method,
                                                           self.authenticated_saml_response,
                                                           self.irrelevant_config)
        # then returns no accounts
        assert accounts == {account_no: account_alias}

    def test_returns_two_accounts_when_two_accounts_are_listed(self):
        # given user with no aws accounts
        account_no = '1'
        account_alias = 'single'
        second_account_no = '2'
        second_account_alias = 'bingle'
        self.authenticated_session.post = lambda *args, **kwargs: _account_page_response([
            _aws_account(account_alias, account_no),
            _aws_account(second_account_alias, second_account_no),
        ])

        # when gets account aliases via fetcher
        accounts = account_aliases_fetcher.account_aliases(self.authenticated_session,
                                                           self.irrelevant_username,
                                                           self.irrelevant_password,
                                                           self.irrelevant_auth_method,
                                                           self.authenticated_saml_response,
                                                           self.irrelevant_config)
        # then returns no accounts
        assert accounts == {account_no: account_alias, second_account_no: second_account_alias}

    def test_returns_accounts_expected_for_real_case_response(self):
        # given response with accounts
        response_text, expected_accounts = self._response_with_expected_aliases()
        self.authenticated_session.post = lambda *args, **kwargs: _account_page_response_text(response_text)

        # when gets account aliases via fetcher
        accounts = account_aliases_fetcher.account_aliases(self.authenticated_session,
                                                           self.irrelevant_username,
                                                           self.irrelevant_password,
                                                           self.irrelevant_auth_method,
                                                           self.authenticated_saml_response,
                                                           self.irrelevant_config)

        # then account numbers matches
        assert accounts.keys() == expected_accounts.keys()

    def test_returns_two_accounts_expected_for_real_case_response(self):
        # given response with accounts
        response_text, expected_accounts = self._response_with_two_expected_aliases()
        self.authenticated_session.post = lambda *args, **kwargs: _account_page_response_text(response_text)

        # when gets account aliases via fetcher
        accounts = account_aliases_fetcher.account_aliases(self.authenticated_session,
                                                           self.irrelevant_username,
                                                           self.irrelevant_password,
                                                           self.irrelevant_auth_method,
                                                           self.authenticated_saml_response,
                                                           self.irrelevant_config)

        # then account numbers matches
        assert accounts.keys() == expected_accounts.keys()

    def test_returns_no_aliases_when_the_call_for_aliases_failed(self):
        # given failed response
        self.authenticated_session.post = lambda *args, **kwargs: _failed_account_page_response()

        # when gets account aliases via fetcher
        accounts = account_aliases_fetcher.account_aliases(self.authenticated_session,
                                                           self.irrelevant_username,
                                                           self.irrelevant_password,
                                                           self.irrelevant_auth_method,
                                                           self.authenticated_saml_response,
                                                           self.irrelevant_config)
        # then there are no aliases
        assert accounts == {}

    def test_returns_full_saml_name_account_when_no_account_alias_is_provided(self):
        # given aws account without alias
        account_no = '123123'
        full_account_name = 'Account: {}'.format(account_no)
        self.authenticated_session.post = lambda *args, **kwargs: _account_page_response([
            _aws_account_without_alias(account_no)
        ])

        # when gets account aliases via fetcher
        accounts = account_aliases_fetcher.account_aliases(self.authenticated_session,
                                                           self.irrelevant_username,
                                                           self.irrelevant_password,
                                                           self.irrelevant_auth_method,
                                                           self.authenticated_saml_response,
                                                           self.irrelevant_config)
        # then uses full account name as the alias
        assert accounts == {account_no: full_account_name}

    def _response_with_expected_aliases(self):
        return u'''
        <html>
        <body>

<div id="container">

  <h1 class="background">Amazon Web Services Login</h1>

  <div id="content">

  <div id="main_error"></div>

  <form id="saml_form" name="saml_form" action="/saml" method="post">
          <input type="hidden" name="RelayState" value="" />
          <input type="hidden" name="SAMLResponse" value="valueofSAMLRESPONSE" />
          <input type="hidden" name="name" value="" />
          <input type="hidden" name="portal" value="" />
          <p style="font-size: 16px; padding-left: 20px;">Select a role:</p>
          <fieldset>
            <div  class="saml-account"> <div onClick="expandCollapse(0);">
              <img id="image0" src="/static/image/down.png" valign="middle"></img>
              <div class="saml-account-name">Account: mydomain-account3 (123456789012)</div>
              </div>
              <hr style="border: 1px solid #ddd;">
              <div id="0" class="saml-account" >
                <div class="saml-role" onClick="checkRadio(this);">
                    <input type="radio" name="roleIndex" value="arn:aws:iam::123456789012:role/ADFS-CloudSearchManager" class="saml-radio" id="arn:aws:iam::123456789012:role/ADFS-CloudSearchManager" />
                    <label for="arn:aws:iam::123456789012:role/ADFS-CloudSearchManager" class="saml-role-description">ADFS-CloudSearchManager</label>
                    <span style="clear: both;"></span>
                </div>
                 </div></div><div  class="saml-account"> <div onClick="expandCollapse(1);">
              <img id="image1" src="/static/image/down.png" valign="middle"></img>
              <div class="saml-account-name">Account: mydomain-account2 (223456789012)</div>
              </div>
              <hr style="border: 1px solid #ddd;">
              <div id="1" class="saml-account" >
                <div class="saml-role" onClick="checkRadio(this);">
                    <input type="radio" name="roleIndex" value="arn:aws:iam::223456789012:role/ADFS-CloudSearchManager" class="saml-radio" id="arn:aws:iam::223456789012:role/ADFS-CloudSearchManager" />
                    <label for="arn:aws:iam::223456789012:role/ADFS-CloudSearchManager" class="saml-role-description">ADFS-CloudSearchManager</label>
                    <span style="clear: both;"></span>
                </div>
                 </div></div><div  class="saml-account"> <div onClick="expandCollapse(2);">
              <img id="image2" src="/static/image/down.png" valign="middle"></img>
              <div class="saml-account-name">Account: mydomain-account1 (323456789012)</div>
              </div>
              <hr style="border: 1px solid #ddd;">
              <div id="2" class="saml-account" >
                <div class="saml-role" onClick="checkRadio(this);">
                    <input type="radio" name="roleIndex" value="arn:aws:iam::323456789012:role/ADFS-Administrator" class="saml-radio" id="arn:aws:iam::323456789012:role/ADFS-Administrator" />
                    <label for="arn:aws:iam::323456789012:role/ADFS-Administrator" class="saml-role-description">ADFS-Administrator</label>
                    <span style="clear: both;"></span>
                </div>

                <div class="saml-role" onClick="checkRadio(this);">
                    <input type="radio" name="roleIndex" value="arn:aws:iam::323456789012:role/ADFS-EMRManager" class="saml-radio" id="arn:aws:iam::323456789012:role/ADFS-EMRManager" />
                    <label for="arn:aws:iam::323456789012:role/ADFS-EMRManager" class="saml-role-description">ADFS-EMRManager</label>
                    <span style="clear: both;"></span>
                </div>
                 </div></div><div  class="saml-account"> <div onClick="expandCollapse(3);">
              <img id="image3" src="/static/image/down.png" valign="middle"></img>
              <div class="saml-account-name">Account: 423456789012</div>
              </div>
              <hr style="border: 1px solid #ddd;">
              <div id="3" class="saml-account" >
                <div class="saml-role" onClick="checkRadio(this);">
                    <input type="radio" name="roleIndex" value="arn:aws:iam::423456789012:role/ADFS-Administrator" class="saml-radio" id="arn:aws:iam::423456789012:role/ADFS-Administrator" />
                    <label for="arn:aws:iam::423456789012:role/ADFS-Administrator" class="saml-role-description">ADFS-Administrator</label>
                    <span style="clear: both;"></span>
                </div>
                 </div></div>
          </fieldset>

          <br>
          <div class="buttoninput" id="input_signin_button">
              <a id="signin_button" class="css3button" href="#" alt="Continue" value="Continue">Sign In</a>
          </div>

  </form>
  </div>
  </body>
  </html>
        ''', {
            '123456789012': 'mydomain-account3',
            '223456789012': 'mydomain-account2',
            '323456789012': 'mydomain-account1',
            '423456789012': '423456789012',
        }

    def _response_with_two_expected_aliases(self):
        return u'''
        <html>
        <body>

<div id="container">

  <h1 class="background">Amazon Web Services Login</h1>

  <div id="content">

  <div id="main_error"></div>

  <form id="saml_form" name="saml_form" action="/saml" method="post">
          <input type="hidden" name="RelayState" value="" />
          <input type="hidden" name="SAMLResponse" value="valueofSAMLRESPONSE" />
          <input type="hidden" name="name" value="" />
          <input type="hidden" name="portal" value="" />
          <p style="font-size: 16px; padding-left: 20px;">Select a role:</p>
<fieldset>
            <div  class="saml-account"> <div onClick="expandCollapse(0);">
              <img id="image0" src="/static/image/down.png" valign="middle"></img>
              <div class="saml-account-name">Account: zefr (123456789012)</div>
              </div>
              <hr style="border: 1px solid #ddd;">
              <div id="0" class="saml-account" >  
                <div class="saml-role" onClick="checkRadio(this);">
                    <input type="radio" name="roleIndex" value="arn:aws:iam::123456789012:role/CORP-ROLE1" class="saml-radio" id="arn:aws:iam::123456789012:role/CORP-ROLE1" />
                    <label for="arn:aws:iam::123456789012:role/CORP-ROLE1" class="saml-role-description">CORP-ROLE1</label>
                    <span style="clear: both;"></span>
                </div>
                
                <div class="saml-role" onClick="checkRadio(this);">
                    <input type="radio" name="roleIndex" value="arn:aws:iam::123456789012:role/CORP-ROLE2" class="saml-radio" id="arn:aws:iam::123456789012:role/CORP-ROLE2" />
                    <label for="arn:aws:iam::123456789012:role/CORP-ROLE2" class="saml-role-description">CORP-ROLE2</label>
                    <span style="clear: both;"></span>
                </div>
                 </div></div>
          </fieldset>
          <br>
          <div class="buttoninput" id="input_signin_button">
              <a id="signin_button" class="css3button" href="#" alt="Continue" value="Continue">Sign In</a>
          </div>

  </form>
  </div>
  </body>
  </html>
        ''', {
            '123456789012': 'zefr',
        }

    def setup_method(self, method):
        self.authenticated_session = type('', (), {})()
        self.irrelevant_auth_method = {}
        self.irrelevant_username = 'irrelevant username'
        self.irrelevant_password = 'irrelevant password'
        self.authenticated_saml_response = 'irrelevant saml response'
        self.irrelevant_config = type('', (), {})()
        self.irrelevant_config.ssl_verification = True



import configparser

import boto3
import botocore
import botocore.exceptions
import botocore.session
import click
from botocore import client
from os import environ
import sys
from . import authenticator
from . import prepare
from . import role_chooser


@click.command()
@click.option(
    '--profile',
    default=lambda: environ.get('AWS_DEFAULT_PROFILE', 'default'),
    help='AWS cli profile that will be authenticated.\n'
         'After successful authentication just use:\n'
         'aws --profile <authenticated profile> <service> ...',
)
@click.option(
    '--region',
    help='The default AWS region that this script will connect\n'
         'to for all API calls',
)
@click.option(
    '--ssl-verification/--no-ssl-verification',
    default=None,
    help='SSL certificate verification: Whether or not strict certificate\n'
         'verification is done, False should only be used for dev/test',
)
@click.option(
    '--adfs-ca-bundle',
    default=None,
    help='Override CA bundle for SSL certificate verification for ADFS server only.',
)
@click.option(
    '--adfs-host',
    help='For the first time for a profile it has to be provided, next time for the same profile\n'
         'it will be loaded from the stored configuration',
)
@click.option(
    '--output-format',
    type=click.Choice(['json', 'text', 'table']),
    help='Output format used by aws cli',
)
@click.option(
    '--provider-id',
    help='Provider ID, e.g urn:amazon:webservices (optional)',
)
@click.option(
    '--s3-signature-version',
    type=click.Choice(['s3v4']),
    help='s3 signature version: Identifies the version of AWS Signature to support for '
         'authenticated requests. Valid values: s3v4',
)
@click.option(
    '--env',
    is_flag=True,
    help='Read username, password from environment variables (username and password).',
)
@click.option(
    '--stdin',
    is_flag=True,
    help='Read username, password from standard input separated by a newline.',
)
@click.option(
    '--authfile',
    help='Read username, password from a local file (optional)',
)
@click.option(
    '--stdout',
    is_flag=True,
    help='Print aws_session_token in json on stdout.',
)
@click.option(
    '--printenv',
    is_flag=True,
    help='Output commands to set AWS_ACCESS_KEY_ID, AWS_SECRET_ACCESS_KEY, AWS_SESSION_TOKEN, AWS_DEFAULT_REGION environmental variables instead of saving them to the aws configuration file.',
)
@click.option(
    '--role-arn',
    help='Predefined role arn to selects, e.g. aws-adfs login --role-arn arn:aws:iam::123456789012:role/YourSpecialRole',
)
@click.option(
    '--session-duration',
    help='Define the amount of seconds you want to establish your STS session, e.g. aws-adfs login --session-duration 3600',
    type=int,
)
@click.option(
    '--assertfile',
    help='Use SAML assertion response from a local file'
)
@click.option(
    '--sspi/--no-sspi',
    default=None,
    help='Whether or not to use Kerberos SSO authentication via SSPI, which may not work in some environments.',
)
@click.option(
    '--u2f-trigger-default/--no-u2f-trigger-default',
    default=None,
    help='Whether or not to also trigger the default authentication method when U2F is available (only works with Duo for now).',
)
def login(
        profile,
        region,
        ssl_verification,
        adfs_ca_bundle,
        adfs_host,
        output_format,
        provider_id,
        s3_signature_version,
        env,
        stdin,
        authfile,
        stdout,
        printenv,
        role_arn,
        session_duration,
        assertfile,
        sspi,
        u2f_trigger_default,
):
    """
    Authenticates an user with active directory credentials
    """
    config = prepare.get_prepared_config(
        profile,
        region,
        ssl_verification,
        adfs_ca_bundle,
        adfs_host,
        output_format,
        provider_id,
        s3_signature_version,
        session_duration,
        sspi,
        u2f_trigger_default,
    )

    _verification_checks(config)

    # Try re-authenticating using an existing ADFS session
    principal_roles, assertion, aws_session_duration = authenticator.authenticate(config, assertfile=assertfile)

    # If we fail to get an assertion, prompt for credentials and try again
    if assertion is None:
        password = None

        if stdin:
            config.adfs_user, password = _stdin_user_credentials()
        elif env:
            config.adfs_user, password = _env_user_credentials()
        elif authfile:
            config.adfs_user, password = _file_user_credentials(config.profile, authfile)

        if not config.adfs_user:
            config.adfs_user = click.prompt(text='Username', type=str, default=config.adfs_user)

        if not password:
            password = click.prompt('Password', type=str, hide_input=True)

        principal_roles, assertion, aws_session_duration = authenticator.authenticate(config, config.adfs_user, password)

        password = '########################################'
        del password

    if(role_arn is not None):
        config.role_arn = role_arn
    principal_arn, config.role_arn = role_chooser.choose_role_to_assume(config, principal_roles)
    if principal_arn is None or config.role_arn is None:
        click.echo('This account does not have access to any roles', err=True)
        exit(-1)

    # Use the assertion to get an AWS STS token using Assume Role with SAML
    # according to the documentation:
    #   http://docs.aws.amazon.com/IAM/latest/UserGuide/id_roles_providers_create_saml_assertions.html
    # This element contains one AttributeValue element that specifies the maximum time that the user
    # can access the AWS Management Console before having to request new temporary credentials.
    # The value is an integer representing the number of seconds, and can be
    # a maximum of 43200 seconds (12 hours). If this attribute is not present,
    # then the maximum session duration defaults to one hour
    # (the default value of the DurationSeconds parameter of the AssumeRoleWithSAML API).
    # To use this attribute, you must configure the SAML provider to provide single sign-on access
    # to the AWS Management Console through the console sign-in web endpoint at
    # https://signin.aws.amazon.com/saml.
    # Note that this attribute extends sessions only to the AWS Management Console.
    # It cannot extend the lifetime of other credentials.
    # However, if it is present in an AssumeRoleWithSAML API call,
    # it can be used to shorten the lifetime of the credentials returned by the call to less than
    # the default of 60 minutes.
    #
    # Note, too, that if a SessionNotOnOrAfter attribute is also defined,
    # then the lesser value of the two attributes, SessionDuration or SessionNotOnOrAfter,
    # establishes the maximum duration of the console session.
    _bind_aws_session_to_chosen_profile(config)
    conn = boto3.client('sts', config=client.Config(signature_version=botocore.UNSIGNED))
    aws_session_token = conn.assume_role_with_saml(
        RoleArn=config.role_arn,
        PrincipalArn=principal_arn,
        SAMLAssertion=assertion,
        DurationSeconds=int(config.session_duration),
    )

    if stdout:
        _emit_json(aws_session_token)
    elif printenv:
        _emit_summary(config, aws_session_duration)
        _print_environment_variables(aws_session_token,config)
    else:
        _store(config, aws_session_token)
        _emit_summary(config, aws_session_duration)


def _bind_aws_session_to_chosen_profile(config):
    try:
        boto3.setup_default_session(profile_name=config.profile)
    except botocore.exceptions.ProfileNotFound:
        pass


def _emit_json(aws_session_token):
    click.echo(
        u"""{{"AccessKeyId": "{}", "SecretAccessKey": "{}", "SessionToken": "{}"}}""".format(
            aws_session_token['Credentials']['AccessKeyId'],
            aws_session_token['Credentials']['SecretAccessKey'],
            aws_session_token['Credentials']['SessionToken']
        )
    )

def _print_environment_variables(aws_session_token,config):
    envcommand = "export"
    if(sys.platform=="win32"):
        envcommand="set"

    click.echo(
        u"""{} AWS_ACCESS_KEY_ID={}""".format(envcommand,aws_session_token['Credentials']['AccessKeyId']))
    click.echo(
        u"""{} AWS_SECRET_ACCESS_KEY={}""".format(envcommand,aws_session_token['Credentials']['SecretAccessKey']))
    click.echo(
        u"""{} AWS_SESSION_TOKEN={}""".format(envcommand,aws_session_token['Credentials']['SessionToken']))
    click.echo(
        u"""{} AWS_DEFAULT_REGION={}""".format(envcommand,config.region))


def _emit_summary(config, session_duration):
    click.echo(
        u"""
        Prepared ADFS configuration as follows:
            * AWS CLI profile                   : '{}'
            * AWS region                        : '{}'
            * Output format                     : '{}'
            * SSL verification of ADFS Server   : '{}'
            * Selected role_arn                 : '{}'
            * ADFS Server                       : '{}'
            * ADFS Session Duration in seconds  : '{}'
            * Provider ID                       : '{}'
            * S3 Signature Version              : '{}'
            * STS Session Duration in seconds   : '{}'
            * SSPI:                             : '{}'
            * U2F and default method            : '{}'
        """.format(
            config.profile,
            config.region,
            config.output_format,
            'ENABLED' if config.ssl_verification else 'DISABLED',
            config.role_arn,
            config.adfs_host,
            session_duration,
            config.provider_id,
            config.s3_signature_version,
            config.session_duration,
            config.sspi,
            config.u2f_trigger_default,
        )
    )


def _file_user_credentials(profile, authfile):
    config = configparser.ConfigParser()

    try:
        if len(config.read(authfile)) == 0:
            raise IOError(authfile)
    except IOError as e:
        print('Auth file ({}) not found'.format(e))
        return None, None

    try:
        username = config.get(profile, "username")
    except configparser.Error:
        print('Failed to read username from auth file, section ({}).'.format(profile))
        username = None

    try:
        password = config.get(profile, "password")
    except configparser.Error:
        print('Failed to read password from auth file, section ({}).'.format(profile))
        password = None

    return username, password


def _env_user_credentials():
    try:
        username = environ['username']
    except:
        print('Failed to read username from env')
        username = None

    try:
        password = environ['password']
    except:
        print('Failed to read password from env')
        password = None

    return username, password


def _stdin_user_credentials():
    stdin = click.get_text_stream('stdin').read()
    stdin_lines = stdin.strip().splitlines()
    try:
        username, password = stdin_lines[:2]
    except ValueError:
        print('Failed to read newline separated username and password from stdin.')
        username = None
        password = None

    return username, password


def _store(config, aws_session_token):
    def store_config(profile, config_location, storer):
        config_file = configparser.RawConfigParser()
        config_file.read(config_location)

        if not config_file.has_section(profile):
            config_file.add_section(profile)

        storer(config_file, profile)

        with open(config_location, 'w+') as f:
            try:
                config_file.write(f)
            finally:
                f.close()

    def credentials_storer(config_file, profile):
        config_file.set(profile, 'aws_access_key_id', aws_session_token['Credentials']['AccessKeyId'])
        config_file.set(profile, 'aws_secret_access_key', aws_session_token['Credentials']['SecretAccessKey'])
        config_file.set(profile, 'aws_session_token', aws_session_token['Credentials']['SessionToken'])
        config_file.set(profile, 'aws_security_token', aws_session_token['Credentials']['SessionToken'])

    def config_storer(config_file, profile):
        config_file.set(profile, 'region', config.region)
        config_file.set(profile, 'output', config.output_format)
        config_file.set(profile, 'adfs_config.ssl_verification', config.ssl_verification)
        config_file.set(profile, 'adfs_config.role_arn', config.role_arn)
        config_file.set(profile, 'adfs_config.adfs_host', config.adfs_host)
        if config.adfs_user:
            config_file.set(profile, 'adfs_config.adfs_user', config.adfs_user)
        if config.s3_signature_version:
            config_file.set(profile, 's3', '\nsignature_version = {}'.format(config.s3_signature_version))
        config_file.set(profile, 'adfs_config.session_duration', config.session_duration)
        config_file.set(profile, 'adfs_config.provider_id', config.provider_id)
        config_file.set(profile, 'adfs_config.sspi', config.sspi)
        config_file.set(profile, 'adfs_config.u2f_trigger_default', config.u2f_trigger_default)

    store_config(config.profile, config.aws_credentials_location, credentials_storer)
    if config.profile == 'default':
        store_config(config.profile, config.aws_config_location, config_storer)
    else:
        store_config('profile {}'.format(config.profile), config.aws_config_location, config_storer)


def _verification_checks(config):
    if not config.adfs_host:
        click.echo('\'--adfs-host\' parameter must be supplied', err=True)
        exit(-1)



import collections
import logging

import click


def choose_role_to_assume(config, principal_roles):
    chosen_principal_arn = None
    chosen_role_arn = None

    principal_roles_emptied = not bool(principal_roles)
    if principal_roles_emptied:
        return chosen_principal_arn, chosen_role_arn

    role_collection = []
    principal_roles = collections.OrderedDict(sorted(principal_roles.items(), key=lambda t: t[0]))
    for account_name in principal_roles.keys():
        roles = principal_roles[account_name]
        for role_arn in roles.keys():
            role_collection.append([roles[role_arn]['principal_arn'], role_arn])

    logging.debug(u'Role arn from config: {}'.format(config.role_arn))

    chosen_principal_role = [role for role in role_collection if config.role_arn == role[1]]

    logging.debug(u'Calculated role collection: {}'.format(role_collection))
    if len(chosen_principal_role) == 1:
        logging.debug(u'Chosen principal role based on previously used role_arn stored in config: {}'
                      .format(chosen_principal_role))
        chosen_principal_arn = chosen_principal_role[0][0]
        chosen_role_arn = chosen_principal_role[0][1]
        return chosen_principal_arn, chosen_role_arn

    if len(role_collection) == 1:
        logging.debug(u'There is only one role to choose')
        chosen_principal_arn = role_collection[0][0]
        chosen_role_arn = role_collection[0][1]
    elif len(role_collection) > 1:
        logging.debug(u'Manual choice')
        click.echo(u'Please choose the role you would like to assume:')
        i = 0
        for account_name in principal_roles.keys():
            roles = principal_roles[account_name]
            click.echo('{}:'.format(account_name))
            for role_arn in roles.keys():
                role_entry = roles[role_arn]
                click.echo('    [ {} -> {} ]: {}'.format(role_entry['name'].ljust(30, ' ' if i % 2 == 0 else '.'), i, role_arn))
                i += 1

        selected_index = click.prompt(text='Selection', type=click.IntRange(0, len(role_collection)))

        chosen_principal_arn = role_collection[selected_index][0]
        chosen_role_arn = role_collection[selected_index][1]

    return chosen_principal_arn, chosen_role_arn



# -*- coding: utf-8 -*-

import logging
import sys

import click

from . import list_profiles
from . import login
from . import reset
from . import _version


def _print_version(ctx, param, value):
    if not value or ctx.resilient_parsing:
        return
    click.echo(_version.get_versions()['version'])
    ctx.exit()


@click.group()
@click.option(
    '--version',
    is_flag=True,
    callback=_print_version,
    expose_value=False,
    is_eager=True,
    help='Show current tool version'
)
@click.option(
    '-v', '--verbose',
    default=False,
    is_flag=True,
    help='Enables debug information on stdout. By default log level is set on ERROR'
)
def cli(verbose):
    log_format = '%(asctime)s [%(module)s %(filename)s:%(funcName)s] ' \
                 '[%(process)d-%(processName)s] [%(thread)d-%(threadName)s] ' \
                 '- %(levelname)s: %(message)s'
    logging.basicConfig(
        format=log_format,
        stream=sys.stderr,
        level=logging.DEBUG if verbose else logging.ERROR,
    )
    if verbose:
        try:
            import http.client as http_client
        except ImportError:
            # Python 2
            import httplib as http_client
        http_client.HTTPConnection.debuglevel = 1

cli.add_command(list_profiles.list_profiles)
cli.add_command(login.login)
cli.add_command(reset.reset)




from aws_adfs import role_chooser
from aws_adfs.role_chooser import click


class TestRoleChooser:

    def test_returns_no_roles_for_empty_list(self):
        # given user without roles
        empty_roles_collection = [{}, None, []]

        for empty_principal_roles in empty_roles_collection:
            # when an user is asked to choose a role
            chosen_principal_arn, chosen_role_arn = role_chooser.choose_role_to_assume(
                config=self.irrelevant_config,
                principal_roles=empty_principal_roles
            )

            # then there are not roles
            assert chosen_principal_arn is None
            assert chosen_role_arn is None

    def test_returns_already_chosen_roles_when_it_is_named_for_an_user(self):
        # given role already chosen by an user
        already_chosen_role_arn = 'already_chosen_role_arn'

        config_with_already_chosen_role = type('', (), {})()
        config_with_already_chosen_role.role_arn = already_chosen_role_arn

        # and roles collection for the user still contains already chosen role
        already_chosen_principal_arn = 'already_chosen_principal_arn'
        roles_collection_with_already_chosen_one = {
            'awesome_account': {
                'irrelevant_arn': {'name': 'irrelevant', 'principal_arn': 'irrelevant'},
                already_chosen_role_arn: {'name': 'irrelevant', 'principal_arn': already_chosen_principal_arn},
            }
        }

        # when an user is asked to choose a role
        chosen_principal_arn, chosen_role_arn = role_chooser.choose_role_to_assume(
            config=config_with_already_chosen_role,
            principal_roles=roles_collection_with_already_chosen_one
        )

        # then returns already chosen role
        assert chosen_principal_arn == already_chosen_principal_arn
        assert chosen_role_arn == already_chosen_role_arn

    def test_returns_one_role_when_only_one_is_available(self):
        # given roles collection for the user contains one available role
        one_available_principal_arn = 'one_available_principal_arn'
        one_available_role_arn = 'one_available_role_arn'
        roles_collection_with_one_available_role = {
            'awesome_account': {
                one_available_role_arn: {'name': 'irrelevant', 'principal_arn': one_available_principal_arn}
            }
        }

        # when an user is asked to choose a role from two available
        chosen_principal_arn, chosen_role_arn = role_chooser.choose_role_to_assume(
            config=self.irrelevant_config,
            principal_roles=roles_collection_with_one_available_role
        )
        # then returns the one role that was assigned to the user
        assert chosen_principal_arn == one_available_principal_arn
        assert chosen_role_arn == one_available_role_arn

    def test_returns_second_role_chosen_by_the_user(self):
        # given roles collection for the user contains two roles
        first_principal_arn = 'first_principal_arn'
        first_role_arn = 'first_role_arn'
        chosen_by_the_user_principal_arn = 'chosen_by_the_user_principal_arn'
        chosen_by_the_user_role_arn = 'chosen_by_the_user_role_arn'
        roles_collection_with_one_available_role = {
            'awesome_account': {
                first_role_arn: {'name': first_role_arn, 'principal_arn': first_principal_arn}
            },
            'second_account': {
                chosen_by_the_user_role_arn: {'name': chosen_by_the_user_role_arn, 'principal_arn': chosen_by_the_user_principal_arn}
            }
        }

        # and the user chosen second role
        click.prompt = lambda **kwargs: 1

        # when an user is asked to choose a role from only one available
        chosen_principal_arn, chosen_role_arn = role_chooser.choose_role_to_assume(
            config=self.irrelevant_config,
            principal_roles=roles_collection_with_one_available_role
        )
        # then returns the role that was chosen by the user
        assert chosen_principal_arn == chosen_by_the_user_principal_arn
        assert chosen_role_arn == chosen_by_the_user_role_arn

    def test_goes_for_roles_choice_when_there_wasnt_any_previously_used_role_in_config(self):
        # given roles collection for the user contains two roles
        first_principal_arn = 'first_principal_arn'
        first_role_arn = 'first_role_arn'
        chosen_by_the_user_principal_arn = 'chosen_by_the_user_principal_arn'
        chosen_by_the_user_role_arn = 'chosen_by_the_user_role_arn'
        roles_collection_with_one_available_role = {
            'awesome_account': {
                first_role_arn: {'name': first_role_arn, 'principal_arn': first_principal_arn}
            },
            'second_account': {
                chosen_by_the_user_role_arn: {'name': chosen_by_the_user_role_arn, 'principal_arn': chosen_by_the_user_principal_arn}
            }
        }

        # and there wasn't any previously used role
        config_without_previously_used_role = type('', (), {})()
        config_without_previously_used_role.role_arn = None

        # and the user chosen second role
        click.prompt = lambda **kwargs: 1

        # when an user is asked to choose a role from only one available
        chosen_principal_arn, chosen_role_arn = role_chooser.choose_role_to_assume(
            config=config_without_previously_used_role,
            principal_roles=roles_collection_with_one_available_role
        )
        # then returns the role that was chosen by the user
        assert chosen_principal_arn == chosen_by_the_user_principal_arn
        assert chosen_role_arn == chosen_by_the_user_role_arn

    def test_lets_user_choose_in_case_of_missing_already_chosen_role_in_current_list(self):
        # given role already chosen by an user
        already_chosen_role_arn = 'already_chosen_role_arn'

        config_with_already_chosen_role = type('', (), {})()
        config_with_already_chosen_role.role_arn = already_chosen_role_arn

        # and roles collection for the user contains two roles without already chosen by the user
        first_principal_arn = 'first_principal_arn'
        first_role_arn = 'first_role_arn'
        chosen_by_the_user_principal_arn = 'chosen_by_the_user_principal_arn'
        chosen_by_the_user_role_arn = 'chosen_by_the_user_role_arn'
        roles_collection_with_one_available_role = {
            'awesome_account': {
                first_role_arn: {'name': first_role_arn, 'principal_arn': first_principal_arn}
            },
            'second_account': {
                chosen_by_the_user_role_arn: {'name': chosen_by_the_user_role_arn, 'principal_arn': chosen_by_the_user_principal_arn}
            }
        }

        # and the user chosen second role
        click.prompt = lambda **kwargs: 1

        # when an user is asked to choose a role from two available
        chosen_principal_arn, chosen_role_arn = role_chooser.choose_role_to_assume(
            config=self.irrelevant_config,
            principal_roles=roles_collection_with_one_available_role
        )
        # then returns the role that was chosen by the user
        assert chosen_principal_arn == chosen_by_the_user_principal_arn
        assert chosen_role_arn == chosen_by_the_user_role_arn

    def test_asks_user_to_choose_a_role(self):
        # given the role is assumed for the first time
        config = type('', (), {})()
        config.role_arn = None

        # and the user chosen second role
        click.prompt = lambda **kwargs: 1

        # when an user is asked to choose a role from two available
        chosen_principal_arn, chosen_role_arn = role_chooser.choose_role_to_assume(
            config=config,
            principal_roles=self._two_roles_grouped_in_one_account()
        )

        # then returns the role that was chosen by the user
        assert chosen_principal_arn is not None
        assert chosen_role_arn is not None

    def _two_roles_grouped_in_one_account(self):
        return {'COMPANY': {'arn:aws:iam::AWSACCOUNT:role/CORP-SuperAdmin': {'principal_arn': 'arn:aws:iam::AWSACCOUNT:saml-provider/ADFS', 'name': 'CORP-SuperAdmin'}, 'arn:aws:iam::AWSACCOUNT:role/CORP-DataScience': {'principal_arn': 'arn:aws:iam::AWSACCOUNT:saml-provider/ADFS', 'name': 'CORP-DataScience'}}}

    def setup_method(self, method):
        self.irrelevant_config = type('', (), {})()
        self.irrelevant_config.role_arn = 'irrelevant_role'



import configparser

import click

from os import environ
from .prepare import create_adfs_default_config


@click.command()
@click.option(
    '--profile',
    default=lambda: environ.get('AWS_DEFAULT_PROFILE', 'default'),
    help='AWS cli profile that will be removed'
)
def reset(profile):
    """
    removes stored profile
    """
    adfs_config = create_adfs_default_config('default')
    _clear_credentials(adfs_config, profile)
    click.echo('Profile: \'{}\' has been wiped out'.format(profile))


def _clear_credentials(config, profile):
    def store_config(config_location, storer):
        config_file = configparser.RawConfigParser()
        config_file.read(config_location)

        if not config_file.has_section(profile):
            config_file.add_section(profile)

        storer(config_file)

        with open(config_location, 'w+') as f:
            try:
                config_file.write(f)
            finally:
                f.close()

    def profile_remover(config_file):
        config_file.remove_section(profile)
        config_file.remove_section('profile {}'.format(profile))

    store_config(config.aws_credentials_location, profile_remover)
    store_config(config.aws_config_location, profile_remover)



import click
import lxml.etree as ET

import logging

from . import roles_assertion_extractor

def extract(html_response, ssl_verification_enabled, session):
    """
    :param html_response: html result of parsing http response
    :param ssl_verification_enabled: bool to enable SSL verification
    :param session: current requests Session object
    :return:
    """

    roles_page_url = _action_url_on_validation_success(html_response)

    click.echo('Additional verification is required. Please check your mobile device', err=True)

    # The POST call in this function hangs until verification is completed
    return _retrieve_roles_page(
        roles_page_url,
        _context(html_response),
        session,
        ssl_verification_enabled
    )

    
def _retrieve_roles_page(roles_page_url, context, session, ssl_verification_enabled):
    response = session.post(
        roles_page_url,
        verify=ssl_verification_enabled,
        allow_redirects=True,
        data={
            'AuthMethod': 'AzureMfaServerAuthentication',
            'Context': context,
        }
    )
    logging.debug(u'''Request:
            * url: {}
            * headers: {}
        Response:
            * status: {}
            * headers: {}
            * body: {}
        '''.format(roles_page_url, response.request.headers, response.status_code, response.headers,
                   response.text))

    if response.status_code != 200:
        raise click.ClickException(
            u'Issues during redirection to aws roles page. The error response {}'.format(
                response
            )
        )

    html_response = ET.fromstring(response.text, ET.HTMLParser())
    return roles_assertion_extractor.extract(html_response)


def _context(html_response):
    context_query = './/input[@id="context"]'
    element = html_response.find(context_query)
    return element.get('value')

def _action_url_on_validation_success(html_response):
    post_url_query = './/form[@id="options"]'
    element = html_response.find(post_url_query)
    return element.get('action')



import base64

import lxml.etree as ET
import pytest

from aws_adfs import roles_assertion_extractor


class TestAssertionExtractor:

    def test_login_error_causes_error_and_exit(self, capsys):
        # given login page after authentication failure
        login_error_page_result = self._a_page_of_authentication_failure()

        # Error is printed and exit is called
        with pytest.raises(SystemExit):
            # when a login page after failed authentication is extracted
            roles_assertion_extractor.extract(login_error_page_result)

        out, err = capsys.readouterr()
        assert err.startswith("Login error: Incorrect user ID or password")

    def test_missing_saml_assertion_causes_to_return_nothing(self):
        # when a returned result page doesn't contain saml (perhaps session expired)
        roles, assertion, _ = roles_assertion_extractor.extract(self._a_page_of_expired_login())

        # the return nothing - perhaps re-authentication is needed
        assert roles is None
        assert assertion is None

    def test_beer_roles_are_extracted(self):
        # when after successful authentication adfs responded with page containing available roles
        roles, assertion, _ = roles_assertion_extractor.extract(self.a_page_of_allowed_beer_roles())

        # then two beer roles are extracted
        assert len(roles) == 2
        assert assertion is not None

    def test_provides_existing_in_response_session_duration(self):
        # adfs has configured aws session duration
        session_duration_configured_in_adfs = 7200

        # when after successful authentican adfs responds with saml response
        # containing session duration
        roles, assertion, extracted_session_duration = roles_assertion_extractor.extract(
            self.a_page_with_saml_containing_session_duration(session_duration_configured_in_adfs)
        )

        # then responded session duration is extracted
        assert extracted_session_duration == session_duration_configured_in_adfs

    def test_provides_default_session_duration_when_it_is_missing_in_response(self):
        # when after successful authentican adfs responds with saml response
        # without session duration setup
        roles, assertion, extracted_session_duration = roles_assertion_extractor.extract(
            self.a_page_with_saml_without_session_duration()
        )

        # then extracted session duration has default value for boto
        assert extracted_session_duration == roles_assertion_extractor.default_session_duration

    @staticmethod
    def a_page_of_allowed_beer_roles():
        assertion = u'''<samlp:Response Consent="urn:oasis:names:tc:SAML:2.0:consent:unspecified" Destination="https://signin.aws.amazon.com/saml" xmlns:samlp="urn:oasis:names:tc:SAML:2.0:protocol">
<Assertion Version="2.0" xmlns="urn:oasis:names:tc:SAML:2.0:assertion">
<AttributeStatement>
  <Attribute Name="https://aws.amazon.com/SAML/Attributes/RoleSessionName">
    <AttributeValue>beer.lover@awesome.company.com</AttributeValue>
  </Attribute>
  <Attribute Name="https://aws.amazon.com/SAML/Attributes/Role">
    <AttributeValue>arn:aws:iam::000000000000:saml-provider/ADFS,arn:aws:iam::000000000000:role/beer-lover</AttributeValue>
    <AttributeValue>arn:aws:iam::000000000000:saml-provider/ADFS,arn:aws:iam::000000000000:role/beer-crafter</AttributeValue>
  </Attribute>
</AttributeStatement>
</Assertion>
</samlp:Response>'''

        encoded_assertion = base64.encodestring(assertion.encode('utf-8')).decode('utf-8')

        return ET.fromstring(
            '''
<html>
  <head>
    <title>Working...</title>
  </head>
  <body>
    <form method="POST" name="hiddenform" action="https://signin.aws.amazon.com:443/saml">
      <input type="hidden" name="SAMLResponse" value="{}" />
    </form>
  </body>
</html>
            '''.format(encoded_assertion),
            ET.HTMLParser(),
        )

    @staticmethod
    def a_page_with_saml_without_session_duration():
        assertion = u'''<samlp:Response Consent="urn:oasis:names:tc:SAML:2.0:consent:unspecified" Destination="https://signin.aws.amazon.com/saml" xmlns:samlp="urn:oasis:names:tc:SAML:2.0:protocol">
<Assertion Version="2.0" xmlns="urn:oasis:names:tc:SAML:2.0:assertion">
<AttributeStatement>
  <Attribute Name="https://aws.amazon.com/SAML/Attributes/RoleSessionName">
    <AttributeValue>beer.lover@awesome.company.com</AttributeValue>
  </Attribute>
  <Attribute Name="https://aws.amazon.com/SAML/Attributes/Role">
    <AttributeValue>arn:aws:iam::000000000000:saml-provider/ADFS,arn:aws:iam::000000000000:role/beer-lover</AttributeValue>
    <AttributeValue>arn:aws:iam::000000000000:saml-provider/ADFS,arn:aws:iam::000000000000:role/beer-crafter</AttributeValue>
  </Attribute>
</AttributeStatement>
</Assertion>
</samlp:Response>'''

        encoded_assertion = base64.encodestring(assertion.encode('utf-8')).decode('utf-8')

        return ET.fromstring(
            '''
<html>
  <head>
    <title>Working...</title>
  </head>
  <body>
    <form method="POST" name="hiddenform" action="https://signin.aws.amazon.com:443/saml">
      <input type="hidden" name="SAMLResponse" value="{}" />
    </form>
  </body>
</html>
            '''.format(encoded_assertion),
            ET.HTMLParser(),
        )

    @staticmethod
    def a_page_with_saml_containing_session_duration(session_duration):
        assertion = u'''<samlp:Response Consent="urn:oasis:names:tc:SAML:2.0:consent:unspecified" Destination="https://signin.aws.amazon.com/saml" xmlns:samlp="urn:oasis:names:tc:SAML:2.0:protocol">
<Assertion Version="2.0" xmlns="urn:oasis:names:tc:SAML:2.0:assertion">
<AttributeStatement>
  <Attribute Name="https://aws.amazon.com/SAML/Attributes/RoleSessionName">
    <AttributeValue>beer.lover@awesome.company.com</AttributeValue>
  </Attribute>
  <Attribute Name="https://aws.amazon.com/SAML/Attributes/Role">
    <AttributeValue>arn:aws:iam::000000000000:saml-provider/ADFS,arn:aws:iam::000000000000:role/beer-lover</AttributeValue>
    <AttributeValue>arn:aws:iam::000000000000:saml-provider/ADFS,arn:aws:iam::000000000000:role/beer-crafter</AttributeValue>
  </Attribute>
  <Attribute Name="https://aws.amazon.com/SAML/Attributes/SessionDuration">
        <AttributeValue>{}</AttributeValue>
      </Attribute>
</AttributeStatement>
</Assertion>
</samlp:Response>'''.format(session_duration)

        encoded_assertion = base64.encodestring(assertion.encode('utf-8')).decode('utf-8')

        return ET.fromstring(
            '''
<html>
  <head>
    <title>Working...</title>
  </head>
  <body>
    <form method="POST" name="hiddenform" action="https://signin.aws.amazon.com:443/saml">
      <input type="hidden" name="SAMLResponse" value="{}" />
    </form>
  </body>
</html>
            '''.format(encoded_assertion),
            ET.HTMLParser(),
        )

    @staticmethod
    def _a_page_of_expired_login():
        return ET.fromstring(
            '''
<!DOCTYPE html>
<html>
  <head>
    <title>Amazon Web Services Sign-In</title>
    <meta name="viewport" content="width=device-width" />
    <meta http-equiv="Content-Type" content="text/html; charset=UTF-8"></head>
    <body>
      <div id="container">
        <h1 class="background">Amazon Web Services Login</h1>
        <div id="content">
          <div id="main_error"></div>
          <form id="saml_form" name="saml_form" action="/saml" method="post">
            <input type="hidden" name="RelayState" value="" />
            <input type="hidden" name="SAMLResponse" value="" />
            <input type="hidden" name="name" value="" />
            <p style="font-size: 16px; padding-left: 20px;">Select a role:</p>
          </div>
        </body>
      </html>
            ''',
            ET.HTMLParser(),
        )

    @staticmethod
    def _a_page_of_authentication_failure():
        return ET.fromstring(
            '''
 <!DOCTYPE html>
<html lang="en-US">
    <head>
        <meta http-equiv="X-UA-Compatible" content="IE=10.000"/>
        <meta name="viewport" content="width=device-width, initial-scale=1.0, maximum-scale=1.0, user-scalable=no"/>
        <meta http-equiv="content-type" content="text/html;charset=UTF-8" />
        <meta http-equiv="cache-control" content="no-cache,no-store"/>
        <meta http-equiv="pragma" content="no-cache"/>
        <meta http-equiv="expires" content="-1"/>
        <meta name='mswebdialog-title' content='Connecting to Awesome Company'/>

        <title>Sign In</title>

    </head>
    <body dir="ltr" class="body">
    <div id="fullPage">
        <div id="brandingWrapper" class="float">
            <div id="branding"></div>
        </div>
        <div id="contentWrapper" class="float">
            <div id="content">
                <div id="header">
                    Awesome Company
                </div>
                <div id="workArea">

    <div id="authArea" class="groupMargin">


    <div id="loginArea">
        <div id="loginMessage" class="groupMargin">Sign in with your organizational account</div>

        <form method="post" id="loginForm" autocomplete="off" novalidate="novalidate" onKeyPress="if (event && event.keyCode == 13) Login.submitLoginRequest();" action="/adfs/ls/idpinitiatedsignon" >
            <div id="error" class="fieldMargin error smallText">
                <label id="errorText" for="">Incorrect user ID or password. Type the correct user ID and password, and try again.</label>
            </div>

            <div id="formsAuthenticationArea">
                <div id="userNameArea">
                    <input id="userNameInput" name="UserName" type="email" value="123123@3223.123" tabindex="1" class="text fullWidth"
                        spellcheck="false" placeholder="someone@example.com" autocomplete="off"/>
                </div>

                <div id="passwordArea">
                     <input id="passwordInput" name="Password" type="password" tabindex="2" class="text fullWidth"
                        placeholder="Password" autocomplete="off"/>
                </div>
                <div id="kmsiArea" style="display:none">
                    <input type="checkbox" name="Kmsi" id="kmsiInput" value="true" tabindex="3" />
                    <label for="kmsiInput">Keep me signed in</label>
                </div>
                <div id="submissionArea" class="submitMargin">
                    <span id="submitButton" class="submit" tabindex="4"
                        onKeyPress="if (event && event.keyCode == 32) Login.submitLoginRequest();"
                        onclick="return Login.submitLoginRequest();">Sign in</span>
                </div>
            </div>
            <input id="optionForms" type="hidden" name="AuthMethod" value="FormsAuthentication"/>
        </form>

             <div id="authOptions">
        <form id="options"  method="post" action="https://adfs.awesome.company.com:443/adfs/ls/idpinitiatedsignon">
            <input id="optionSelection" type="hidden" name="AuthMethod" />
            <div class='groupMargin'></div>
        </form>
      </div>

        <div id="introduction" class="groupMargin">

        </div>

    </div>

    </div>

                </div>
                <div id="footerPlaceholder"></div>
            </div>
            <div id="footer">
                <div id="footerLinks" class="floatReverse">
                     <div><span id="copyright">&#169; 2013 Microsoft</span></div>
                </div>
            </div>
        </div>
    </div>
    </body>
</html>
            ''',
            ET.HTMLParser(),
        )
